@@ -29,15 +29,11 @@ class MongodbToolboxError(Exception):
     pass
 
 
-def dummy_stats_func(_key: str, **_kwargs: Any) -> None:
-    pass
-
-
 def bulk_write(
     db: Database[RawBSONDocument],
     colname: str,
     ops: list[DatabaseOperation],
-    stats_callback: StatsCallback = dummy_stats_func,
+    stats_callback: Optional[StatsCallback] = None,
 ) -> BulkWriteResult:
     """
     Apply multiple data operations to the collection using mongodb bulk interface.
@@ -47,24 +43,28 @@ def bulk_write(
     :param ops: list of mongodb operations
     :param stats_callback: callback to track statistics
     """
-    stats_callback("bulk-write-%s-ops" % colname, len(ops))
+    if stats_callback:
+        stats_callback("bulk-write-%s-ops" % colname, len(ops))
     bulk_res = db[colname].bulk_write(ops, ordered=False)
     for stats_key, result_key in [
         ("inserted", "nInserted"),
         ("upserted", "nUpserted"),
         ("modified", "nModified"),
     ]:
-        stats_callback(
-            "{}-{}".format(colname, stats_key),
-            bulk_res.bulk_api_result[result_key],
-        )
+        if stats_callback:
+            stats_callback(
+                "bulk-write-{}-{}".format(colname, stats_key),
+                bulk_res.bulk_api_result[result_key],
+            )
     return bulk_res
 
 
 class BulkWriter:
     """
-    Class to collect mongodb operations and execute them with bulk interface
-    when number of pending operations reaches threshold.
+    Class to collect mongodb operations and execute them with bulk interface.
+
+    Actual bulk write happens when number of pending operations reaches
+    defined threshold.
     """
 
     def __init__(
@@ -72,12 +72,15 @@ class BulkWriter:
         db: Database[RawBSONDocument],
         colname: str,
         bulk_size: int = 100,
-        stats_callback: StatsCallback = dummy_stats_func,
+        stats_callback: Optional[StatsCallback] = None,
     ) -> None:
         """
+        Build BulkWriter instance.
+
         :param db: database object
         :param colname: name of collection to apply operations to
-        :param bulk_size: number of operations to store before run them with bulk interface
+        :param bulk_size: number of operations to store before run them with
+            bulk interface
         :param stats_callback: callback to track statistics
         """
         self.db = db
@@ -94,7 +97,7 @@ class BulkWriter:
         return res  # noqa: R504
 
     def update_one(self, *args: Any, **kwargs: Any) -> Optional[BulkWriteResult]:
-        """Add new UpdateOne operation to list of pending operations
+        r"""Add new UpdateOne operation to list of pending operations.
 
         :param \*args: goes directly to UpdateOne constructor
         :param \**kwargs: goes directly to UpdateOne constructor
@@ -108,7 +111,7 @@ class BulkWriter:
         return None
 
     def insert_one(self, *args: Any, **kwargs: Any) -> Optional[BulkWriteResult]:
-        """Add new InsertOne operation to list of pending operations
+        r"""Add new InsertOne operation to list of pending operations.
 
         :param \*args: goes directly to UpdateOne constructor
         :param \**kwargs: goes directly to UpdateOne constructor
@@ -122,7 +125,7 @@ class BulkWriter:
         return None
 
     def flush(self) -> Optional[BulkWriteResult]:
-        """Run all pending operations"""
+        """Run all pending operations."""
         if self.ops:
             return self._write_ops()
         return None
@@ -179,14 +182,15 @@ def only_dup_key_errors(err: BulkWriteError) -> bool:
     )
 
 
-def bulk_insert_dup_retok(  # noqa: CCR001
+def bulk_insert_dup_retok(  # noqa: CCR001, C901
     db: Database[RawBSONDocument],
     colname: str,
     ops: list[InsertOne[Any]],
     dup_key: Union[str, list[str]],
-    stats_callback: StatsCallback = dummy_stats_func,
+    stats_callback: Optional[StatsCallback] = None,
 ) -> list[Any]:
-    stats_callback("bulk-insert-dup-retok-%s-ops" % colname, len(ops))
+    if stats_callback:
+        stats_callback("bulk-insert-dup-retok-%s-ops" % colname, len(ops))
     if isinstance(dup_key, str):
         dup_key = [dup_key]
     slots = set()
@@ -220,21 +224,24 @@ def bulk_insert_dup_retok(  # noqa: CCR001
             tuple(err["op"][x] for x in dup_key) for err in ex.details["writeErrors"]
         }
         ret_slots = list(slots - error_slots)
-        stats_callback("%s-inserted" % colname, len(ret_slots))
+        if stats_callback:
+            stats_callback("bulk-write-%s-inserted" % colname, len(ret_slots))
         return ret_slots
     else:
-        stats_callback("%s-inserted" % colname, len(slots))
+        if stats_callback:
+            stats_callback("bulk-write-%s-inserted" % colname, len(slots))
         return list(slots)
 
 
-def bulk_insert_dup(
+def bulk_insert_dup(  # noqa: CCR001
     db: Database[RawBSONDocument],
     colname: str,
     ops: list[InsertOne[Any]],
-    stats_callback: StatsCallback = dummy_stats_func,
+    stats_callback: Optional[StatsCallback] = None,
 ) -> None:
     """Write multiple insert operations ignoring all duplicate key errors."""
-    stats_callback("bulk-insert-dup-%s-ops" % colname, len(ops))
+    if stats_callback:
+        stats_callback("bulk-insert-dup-%s-ops" % colname, len(ops))
     for op in ops:
         if not isinstance(op, InsertOne):
             raise MongodbToolboxError(
@@ -246,9 +253,11 @@ def bulk_insert_dup(
     except BulkWriteError as ex:
         if not only_dup_key_errors(ex):
             raise
-        stats_callback(
-            "%s-inserted" % colname,
-            len(ops) - len(ex.details["writeErrors"]),
-        )
+        if stats_callback:
+            stats_callback(
+                "bulk-write-%s-inserted" % colname,
+                len(ops) - len(ex.details["writeErrors"]),
+            )
     else:
-        stats_callback("%s-inserted" % colname, len(ops))
+        if stats_callback:
+            stats_callback("bulk-write-%s-inserted" % colname, len(ops))
